﻿using System;

namespace Utilities
{
	/// <summary>
	///     This is a helper object that represents a mathematical interval.
	///     <para>
	///         <code>
	///           This code is published using the MIT license:
	///          
	///           Copyright (c) 2012 Unterrainer Informatik OG
	///           Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
	///           documentation files (the "Software"), to deal in the Software without restriction, including without limitation
	///           the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software,
	///           and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
	///           The above copyright notice and this permission notice shall be included in all copies or substantial portions
	///           of the Software.
	///           THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED
	///           TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
	///           THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
	///           CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
	///           DEALINGS IN THE SOFTWARE.
	///  		</code>
	///         In plain english:
	///         You may use it, free of charge, in all possible ways, even in commercial projects.
	///         The use of this software is your responsibility, as well as the resulting consequences.
	///         It is not copy-left and I will not even sue you, if you're removing this notice and tell everybody that this
	///         code was your idea. Promised.
	///         If you feel like it you may credit me or our site http://www.unterrainer.info, or both, but you don't have to.
	///     </para>
	/// </summary>
	/// <typeparam name="T">The type of the interval.</typeparam>
	public struct Interval<T> where T : IComparable<T>
	{
		private T _maximalValue;
		private T _minimalValue;
		private bool _isMinValueExclusive;
		private bool _isMaxValueExclusive;

		/// <summary>
		///     Initializes a new instance of the <see cref="Interval&lt;T&gt;" /> structure.
		/// </summary>
		/// <param name="min">The minimal value.</param>
		/// <param name="max">The maximal value.</param>
		/// <param name="isMinValueExclusive">
		///     if set to <c>true</c> the lower bound is exclusive, otherwise its inclusive.
		/// </param>
		/// <param name="isMaxValueExclusive">
		///     if set to <c>true</c> the upper bound is exclusive, otherwise its inclusive.
		/// </param>
		/// <exception cref="ArgumentException">Min has to be smaller than or equal to max.</exception>
		public Interval(T min, T max, bool isMinValueExclusive, bool isMaxValueExclusive)
		{
			_isMinValueExclusive = isMinValueExclusive;
			_isMaxValueExclusive = isMaxValueExclusive;
			// min > max OR max < min
			if (min.CompareTo(max) > 0 || max.CompareTo(min) < 0)
			{
				throw new ArgumentException("Min has to be smaller than or equal to max.");
			}
			_minimalValue = min;
			_maximalValue = max;
		}

		/// <summary>
		///     Initializes a new instance of the <see cref="Interval&lt;T&gt;" /> structure.
		///     The default way this structure treats the bounds is set to inclusive.
		/// </summary>
		/// <param name="min">The minimal value.</param>
		/// <param name="max">The maximal value.</param>
		/// <exception cref="ArgumentException">Min has to be smaller than or equal to max.</exception>
		public Interval(T min, T max)
		{
			_isMinValueExclusive = false;
			_isMaxValueExclusive = false;
			// min > max OR max < min
			if (min.CompareTo(max) > 0 || max.CompareTo(min) < 0)
			{
				throw new ArgumentException("Min has to be smaller than or equal to max.");
			}
			_minimalValue = min;
			_maximalValue = max;
		}

		/// <summary>
		///     Gets or sets the minimal value. Adjusts the max-value as well if it
		///     should currently be smaller than the given value.
		/// </summary>
		/// <value>The minimal value.</value>
		public T Min
		{
			get { return _minimalValue; }
			set
			{
				if (value.CompareTo(_maximalValue) > 0)
				{
					_maximalValue = value;
				}
				_minimalValue = value;
			}
		}

		/// <summary>
		///     Gets or sets the maximal value. Adjusts the min-value as well if it
		///     should currently be greater than the given value.
		/// </summary>
		/// <value>The maximal value.</value>
		public T Max
		{
			get { return _maximalValue; }
			set
			{
				if (value.CompareTo(_minimalValue) < 0)
				{
					_minimalValue = value;
				}
				_maximalValue = value;
			}
		}

		/// <summary>
		///     Gets or sets a value indicating whether the min value is inclusive
		///     (within the interval) or not.
		/// </summary>
		/// <value>
		///     if set to <c>true</c> the lower bound is exclusive, otherwise its inclusive.
		/// </value>
		public bool IsMinValueExclusive
		{
			get { return _isMinValueExclusive; }
			set { _isMinValueExclusive = value; }
		}

		/// <summary>
		///     Gets or sets a value indicating whether the max value is inclusive
		///     (within the interval) or not.
		/// </summary>
		/// <value>
		///     if set to <c>true</c> the upper bound is exclusive, otherwise its inclusive.
		/// </value>
		public bool IsMaxValueExclusive
		{
			get { return _isMaxValueExclusive; }
			set { _isMaxValueExclusive = value; }
		}

		/// <summary>
		///     Determines whether a specified value is in between the intervals
		///     boundaries.
		/// </summary>
		/// <param name="value">The value.</param>
		/// <returns>
		///     <c>true</c> if the given value is in between the specified value; otherwise,
		///     <c>false</c>.
		/// </returns>
		public bool IsInBetween(T value)
		{
			return IsInBetween(value, _isMinValueExclusive, _isMaxValueExclusive);
		}

		/// <summary>
		///     Determines whether a specified value is in between the intervals
		///     boundaries.
		/// </summary>
		/// <param name="value">The value.</param>
		/// <param name="isMinValueExclusive">
		///     if set to <c>true</c> treats the lower boundary as exclusive.
		/// </param>
		/// <param name="isMaxValueExclusive">
		///     if set to <c>true</c> treats the upper boundary as exclusive.
		/// </param>
		/// <returns>
		///     <c>true</c> if the given value is in between the specified value; otherwise,
		///     <c>false</c>.
		/// </returns>
		public bool IsInBetween(T value, bool isMinValueExclusive, bool isMaxValueExclusive)
		{
			bool isGreaterThanMin;
			bool isSmallerThanMax;
			if (isMinValueExclusive)
			{
				isGreaterThanMin = value.CompareTo(Min) > 0;
			}
			else
			{
				isGreaterThanMin = value.CompareTo(Min) >= 0;
			}

			if (!isGreaterThanMin)
			{
				return false;
			}

			if (isMaxValueExclusive)
			{
				isSmallerThanMax = value.CompareTo(Max) < 0;
			}
			else
			{
				isSmallerThanMax = value.CompareTo(Max) <= 0;
			}
			return isSmallerThanMax;
		}
	}
}